#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>

#include "jcm_fetools.h"


#if defined(_WIN32) 
  #define _SEP_ "\\"
#else
  #define _SEP_ "/"
#endif


std::string GetTableContent(int handle)
{
  std::stringstream pStream;

  const char* title;
  JCMTableGetTitle(&title,handle);
  int nColumns;
  JCMTableNColumns(&nColumns,handle);

  pStream<<"Table \""<<title<<"\":"<<std::endl;
  
  for (int iC=0;iC<nColumns;iC++)
  {
    const char* colName;
    JCMTableGetColumnName(&colName,handle,iC);
    pStream<<"Column "<<iC+1<<": "<<colName<<std::endl;
  }

  int nRows;
  JCMTableNRows(&nRows,handle);
  pStream<<"Content:"<<std::endl;
  for (int iR=0;iR<nRows;iR++)
  {
    pStream<<"Row "<<iR<<": ";
    for (int iC=0;iC<nColumns;iC++)
    {
      char type;
      JCMTableGetColumnType(&type,handle,iC);
      switch (type)
      {
      case 'I':
      {
        int entry;
        JCMTableGetIntEntry(&entry,handle,iR,iC);
        pStream<<entry;
        break;
      }
      case 'R':
      {
        double entry;
        JCMTableGetDoubleEntry(&entry,handle,iR,iC);
        pStream<<entry;
        break;
      }
      case 'C':
      {
        doublecomplex entry;
        JCMTableGetDoubleComplexEntry(&entry,handle,iR,iC);
        pStream<<entry.real<<"+i*"<<entry.imag;
        break;
      }
      default:
        break;
      }
      if (iC<nColumns-1)
        pStream<<",";
      pStream<<"\t";
    }
    pStream<<std::endl;
  }
  return pStream.str();
}

std::string GetPinboardStatus()
{
  int nHandles;
  JCMPinboardSize(&nHandles);
  int* handles=new int[nHandles];
  JCMHandleList(&handles);

  std::stringstream pStream;
  pStream<<"Pinboard status:"<<std::endl;
  for (int iH=0;iH<nHandles;iH++)
  {
    int handle=handles[iH];
    const char* type;
    JCMWhatIs(&type,handle);
    const char* source;
    JCMSource(&source,handle);
    int nRef;
    JCMHandleReferenceCount(&nRef,handle);
    pStream<<"\""<<type<<"\" object "<<"<"<<source<<">  (NReferences="<<(nRef-1)<<")\n";
    JCMReleaseHandle(handle);
  }
  delete[] handles;
  return pStream.str();
}



int ReadJCMInputFile(const std::string& path,
                     std::vector<std::string>& inputFiles,
                     std::vector<std::string>& inputContent)
{
  inputFiles.push_back("layout.jcm");
  inputFiles.push_back("project.jcmp");
  inputFiles.push_back("materials.jcm");
  inputFiles.push_back("sources.jcm");
  for (size_t iF=0;iF<inputFiles.size();iF++)
  {
    std::ifstream ifs((path+_SEP_+inputFiles[iF]).c_str());
    if (!ifs.good())
    {
      std::cerr<<"*** File \""<<inputFiles[iF]<<"\" not present"<<std::endl;
      return 1;
    }
    std::string fileContent((std::istreambuf_iterator<char>(ifs)),
                            (std::istreambuf_iterator<char>()));
    inputContent.push_back(fileContent);
  }
  return 0;
}



int main(int argc, char* argv[])
{
  if (argc==1) 
  {
    std::cout<<"*** Please provide path to JCM project directory"<<std::endl;
    return 1;
  }

  //initialization of JCMsuite
  bool initializeMPI=false;
  JCMInitialize(true,initializeMPI);
  JCMGeoInitialize(initializeMPI);
  JCMSetNumThreads(1);

  //read input files
  std::string path(argv[1]);
  std::vector<std::string> inputFiles;
  std::vector<std::string> inputContent;
  if (ReadJCMInputFile(path,inputFiles,inputContent))
    return 1;

  //input files are loaded onto folder on pinboard
  std::string pathPinboard=path+_SEP_+"virtual";

  //create mesh
  int interrupt=0;
  const char* layoutString=inputContent[0].c_str();
  std::string gridFileName=pathPinboard+_SEP_+"grid.jcm";

  int gridHandle;
  if (JCMCreateGridV2(&gridHandle,
                      gridFileName.c_str(),
                      layoutString,
                      &interrupt))
  {
    std::cerr<<"*** Mesh creation failed"<<std::endl;
    return 1;    
  }

  //upload JCMsuite input files to pinboard
  std::vector<int> inputFileHandles;
  for (size_t iF=1;iF<inputFiles.size();iF++)
  {
    int handleLoc;
    std::string filePath=pathPinboard+_SEP_+inputFiles[iF];
    if (JCMSetDataTreeHandle(&handleLoc,inputContent[iF].c_str(),filePath.c_str()))
    {
      std::cerr<<"*** File \""<<inputFiles[iF]<<"\" could not be created on pinboard"<<std::endl;
      return 1;
    }
    inputFileHandles.push_back(handleLoc);
  }

  //solve project
  std::vector<int> resultHandles_(3);
  for (size_t iR=0;iR<resultHandles_.size();iR++)
    resultHandles_[iR]=0;
  int* resultHandles=&resultHandles_[0];
  int nResultHandles;
  int projectHandle=inputFileHandles[0];
  if (JCMSolve(&resultHandles,&nResultHandles,projectHandle,&interrupt))
  {
    std::cerr<<"*** Solving failed"<<std::endl;
    return 1;
  }
  std::cout<<GetPinboardStatus()<<std::endl;


  //read results from pinboard
  for (int iR=0;iR<nResultHandles;iR++)
  {
    const char* fileName_;
    JCMSource(&fileName_,resultHandles_[iR]);

    std::string fileName(fileName_);
    if (fileName.compare(fileName.length()-6,6,"ft.jcm")==0)
      std::cout<<GetTableContent(resultHandles_[iR])<<std::endl;
  }

  //clean up handles
  JCMReleaseHandle(gridHandle);
  for (int iR=0;iR<nResultHandles;iR++)
    JCMReleaseHandle(resultHandles_[iR]);
  for (size_t iR=0;iR<inputFileHandles.size();iR++)
    JCMReleaseHandle(inputFileHandles[iR]);

  JCMFinalize(false);
  JCMGeoFinalize(false);
}
